/**
	$Id $

	Copyright (C) 2014 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.extensions.skin;

import java.util.HashMap;
import java.util.Map;

/**
	Remaps image URLs. Used by {@link SkinAction}:s to
	remap images to other versions.
	
	@author nicklas
	@since 3.4
*/
public class ImageRemapper
{
	
	private final Map<String, String> map;
	
	/**
		Creates a new image remapper. 
	*/
	public ImageRemapper()
	{
		this.map = new HashMap<String, String>();
	}
	
	/**
		Remap a source image to a destination image. The paths
		should be relative the context root and start with /.
		Null values are not ignored. Mappings that create a circular 
		reference are ignored.
		@return TRUE if the mapping was accepted, FALSE if not
	*/
	public boolean map(String src, String dest)
	{
		if (src == null || dest == null) return false;
		if (src.equals(dest)) return false;
		
		// Check if this would create a circular reference
		String circular = map.get(dest);
		int i = 0;
		while (circular != null && i < 100) // 100=safety against infinite recursion
		{
			i++;
			if (circular.equals(src)) return false;
			circular = map.get(circular);
		}
		
		map.put(src, dest);
		return true;
	}
	
	/**
		Get the remapped image for the given source image. The path
		should be relative the context root and start with /.
		@return The remapped path or null if the image is not remapped
	*/
	public String get(String src)
	{
		String dest = map.get(src);
		if (dest != null)
		{
			// Follow multiple re-mappings to the end
			String nextDest = map.get(dest);
			int i = 0;
			while (nextDest != null && i < 100) // 100=safety against infinite recursion
			{
				i++;
				dest = nextDest;
				nextDest = map.get(dest);
			}
		}
		
		return dest;
	}
	
	/**
		Get the number of remapped images.
	*/
	public int size()
	{
		return map.size();
	}
	
	@Override
	public String toString() 
	{
		return map.toString();
	}
	
}
