/*
	$Id$

	Copyright (C) 2014 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.servlet;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import net.sf.basedb.clients.web.extensions.ExtensionsControl;
import net.sf.basedb.util.Values;

import java.io.IOException;
import java.util.HashSet;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
	This is a filter that sets the "Content-Security-Policy" header
	in the http response. See http://base.thep.lu.se/ticket/1712
	for more information.

	@author Nicklas
	@since 3.3
	@base.modified $Date$
*/
public final class ContentSecurityPolicyFilter
	implements Filter
{

	private static boolean UNSAFE_INLINE_SCRIPTS_ARE_ALLOWED = true;
	
	// Contain the file name of all JAR files declared to have unsafe resources
	private static final Set<String> unsafeResources = new HashSet<String>();
	
	private static final Pattern PATH_MATCH = 
			Pattern.compile("(" + ExtensionsControl.RESOURCES_URL + "/([^/]+)/(.*))");

	/**
		Checks if the current policy allows 'unsafe-inline' for the 'script-src'
		attribute. If not we need to check extensions and display a warning to the
		admin that the extension may not work.
	*/
	public static boolean isUnsafeInlineScriptsAllowed()
	{
		return UNSAFE_INLINE_SCRIPTS_ARE_ALLOWED;
	}
	
	/**
		Set a flag indicating if resources from the given extension JAR file are
		should use the regular policy or the 'unsafe-resources' policy. 
	*/
	public static void setSafeResources(String jarName, boolean safeResources)
	{
		synchronized (unsafeResources)
		{
			if (safeResources)
			{
				unsafeResources.remove(jarName);
			}
			else
			{
				unsafeResources.add(jarName);
			}
		}
	}
	
	private String policy = null;
	private String unsafeResourcesPolicy = null;
	private boolean reportOnly = false;
	private String cspHeader;
	
	
	public ContentSecurityPolicyFilter()
	{}
	
	@Override
	public void init(FilterConfig config)
	{
		policy = config.getInitParameter("policy");
		if (policy != null)
		{
			policy = addReportUriIfNeeded(policy, config);
			
			// Check if the policy for 'script-src' contains 'unsafe-inline'
			UNSAFE_INLINE_SCRIPTS_ARE_ALLOWED = false;
			for (String p : policy.split(";"))
			{
				if (p.contains("script-src"))
				{
					UNSAFE_INLINE_SCRIPTS_ARE_ALLOWED = p.contains("'unsafe-inline'");
					break;
				}
				if (p.contains("default-src"))
				{
					UNSAFE_INLINE_SCRIPTS_ARE_ALLOWED = p.contains("'unsafe-inline'");
				}
			}
		}
		
		unsafeResourcesPolicy = config.getInitParameter("unsafe-resources-policy");
		if (unsafeResourcesPolicy != null)
		{
			unsafeResourcesPolicy = addReportUriIfNeeded(unsafeResourcesPolicy, config);
		}
		
		reportOnly = Values.getBoolean(config.getInitParameter("report-only"));
		cspHeader = reportOnly ? "Content-Security-Policy-Report-Only" : "Content-Security-Policy";
	}
	
	@Override
	public void destroy()
	{}
	
	@Override
	public void doFilter(ServletRequest request, ServletResponse response, FilterChain chain)
		throws IOException, ServletException
	{
		HttpServletRequest req = (HttpServletRequest)request;
		
		String usePolicy = policy;
		if (req.getServletPath().startsWith(ExtensionsControl.RESOURCES_URL))
		{
			Matcher m = PATH_MATCH.matcher(req.getServletPath());
			if (m.matches())
			{
				if (unsafeResources.contains(m.group(2)))
				{
					usePolicy = unsafeResourcesPolicy;
				}
			}
		}
		
		if (policy != null) 
		{
			HttpServletResponse resp = (HttpServletResponse) response;
			resp.setHeader(cspHeader, usePolicy);
		}
		
		chain.doFilter(request, response);
	}
	
	private String addReportUriIfNeeded(String policy, FilterConfig config)
	{
		if (policy.indexOf("report-uri") == -1)
		{
			// Report to /csp-report.jsp unless a report-uri has already been specified
			String contextPath = config.getServletContext().getContextPath();
			if (!policy.endsWith(";")) policy += ";";
			policy += "report-uri " + contextPath + "/csp-report;";
		}
		return policy;
	}
}
