/*
	$Id: DynamicReporterQuery.java 6127 2012-09-14 07:28:35Z nicklas $
	
	Copyright (C) 2007 Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.HashSet;
import java.util.Set;

import net.sf.basedb.core.query.AutoJoiner;
import net.sf.basedb.core.query.JoinType;
import net.sf.basedb.core.query.ReporterListExpression;
import net.sf.basedb.core.query.Select;

/**
	Represents a query for reporter data in a "dynamic" way. The main difference 
	from a client applications point of view, is that it is possible to 
	specify which columns that should be selected. In a techinical point of
	view the difference is that we are using SQL instead of HQl.
	
	@author Nicklas
	@version 2.4
	@see Reporter#getDynamicQuery()
	@base.modified $Date: 2012-09-14 09:28:35 +0200 (fr, 14 sep 2012) $
*/
public class DynamicReporterQuery
	extends AbstractSqlQuery
	implements ClonableReporterQuery
{

	private final String rootAlias;
	private final VirtualDb virtualDb;
	private final boolean hasClonedReporters;
	private boolean useClonedReporters;
	private Set<Object> joinedItems;

	/**
		Create a new dynamic query for reporter data.		
	*/
	DynamicReporterQuery()
	{
		super(RealTable.REPORTERS.getQualifiedTableName());
		this.rootAlias = RealTable.REPORTERS.getAlias();
		joinedItems = new HashSet<Object>();
		this.virtualDb = null;
		this.hasClonedReporters = false;
		setAutoJoinType(JoinType.LEFT);
	}
	
	/**
		Create a new dynamic query for reporter data using cloned
		reporter information.
		@since 3.1	
	*/
	DynamicReporterQuery(VirtualDb vdb)
	{
		super(VirtualTable.CLONED_REPORTERS.getQualifiedTableName(vdb));
		this.rootAlias = VirtualTable.CLONED_REPORTERS.getAlias();
		this.virtualDb = vdb;
		this.hasClonedReporters = virtualDb.hasClonedReporters();
		this.useClonedReporters = hasClonedReporters;
		joinedItems = new HashSet<Object>();
		setAutoJoinType(JoinType.LEFT);
	}
	
	/*
		From the Query interface
		-------------------------------------------
	*/
	/**
		The alias of the item that is returned by this query.
		@see Item#getAlias()
	*/
	@Override
	public String getRootAlias()
	{
		return HibernateUtil.quote(rootAlias);
	}
	@Override
	public boolean isReadonly()
	{
		return false;
	}
	// -------------------------------------------

	/*
		From the AbstractSqlQuery class
		-------------------------------------------
	*/
	/**
		No default selection.
	*/
	@Override
	Select[] getDefaultSelects()
	{
		return null;
	}
	// -------------------------------------------

	/**
		Check if the virtual database that this query is using has 
		cloned reporter annotations or not.
		@since 3.1
	*/
	@Override
	public boolean hasClonedReporters()
	{
		return hasClonedReporters;
	}
	
	/**
		Checks if this query is using cloned reporter or not by default.
		The default setting is to use cloned reporters if they are available
		in the experiment.
		
		@return TRUE if this query is using cloned reporters, FALSE if not
		@since 3.1
	*/
	@Override
	public boolean useClonedReporters()
	{
		return useClonedReporters;
	}
	
	/**
		Set if cloned reporters should be used by default or not.
		@param useClonedReporters TRUE to use cloned reporters by default, FALSE
			to use the master reporters
		@since 3.1
	*/
	@Override
	public void setUseClonedReporters(boolean useClonedReporters)
	{
		this.useClonedReporters = hasClonedReporters && useClonedReporters;
	}
	
	/**
		Get the virtual database that this query is using.
		@return A virtual database item (can be null if hasClonedReporter is false)
		@since 3.1
	*/
	@Override
	public VirtualDb getVirtualDb()
	{
		return virtualDb;
	}
	// -------------------------------------------

	/**
	  	Makes a join with a <code>ReporterList</code>
	  	@param reporterList To do the join on.
	  	@param joinType The join type to use. 
		@see net.sf.basedb.core.query.JoinType
		@since 3.0
	*/
	public void joinReporterList(ReporterList reporterList, JoinType joinType)
	{
		if (joinType == null) throw new InvalidUseOfNullException("joinType");
		if (reporterList == null) throw new InvalidUseOfNullException("reporterList");
		if (!joinedItems.contains(reporterList))
		{
			if (useClonedReporters())
			{
				join(new ReporterListJoin(reporterList, joinType, VirtualTable.CLONED_REPORTERS, VirtualColumn.REPORTER_ID_PK));
			}
			else
			{
				join(new ReporterListJoin(reporterList, joinType, RealTable.REPORTERS, "id"));
			}
			joinedItems.add(reporterList);
		}
	}
	
	/**
		Join the master reporter table into this query. This is ignored unless the
		query rooted at the cloned reporters table.
		@param joinType The join type to use 
		@since 3.1
	*/
	public void joinMasterReporters(JoinType joinType)
	{
		if (joinType == null) throw new InvalidUseOfNullException("joinType");
		if (hasClonedReporters() && !joinedItems.contains(RealTable.REPORTERS))
		{
			join(new RealJoin(joinType, VirtualTable.CLONED_REPORTERS, RealTable.REPORTERS, VirtualColumn.REPORTER_ID_PK));
			joinedItems.add(RealTable.REPORTERS);
		}
	}
	
	/**
		Specify the join type of automatic joins. The default join type is
		{@link JoinType#LEFT}. This setting doesn't affect joins that has been made
		directly to the query.
		@param joinType The type of the automatic joins
		@since 3.0
	*/
	public void setAutoJoinType(JoinType joinType)
	{
		addAutoJoiner(getReporterListJoiner(joinType));
		addAutoJoiner(getMasterReporterJoiner(joinType));
	}

	/**
		Get a joiner instance for reporter lists for the specified join type.
		@param joinType The type of join
		@since 3.0
	*/
	public static ReporterListJoiner getReporterListJoiner(JoinType joinType)
	{
		return new ReporterListJoiner(joinType);
	}
	
	/**
		Get a joiner instance for joining the master reporter table
		when the root query is against the cloned reporters table.
		@param joinType The type of join
		@since 3.1
	*/
	public static MasterReporterJoiner getMasterReporterJoiner(JoinType joinType)
	{
		return new MasterReporterJoiner(joinType);
	}
	
	/**
		Auto joiner for reporter lists.
		@since 3.0
	*/
	static class ReporterListJoiner
		implements AutoJoiner<DynamicReporterQuery, ReporterListExpression>
	{
		private final JoinType joinType;
	
		private ReporterListJoiner(JoinType joinType)
		{
			this.joinType = joinType;
		}
		
		/*
			From the AutoJoiner interface
			-------------------------------------------
		*/
		@Override
		public Class<DynamicReporterQuery> getQueryClass()
		{
			return DynamicReporterQuery.class;
		}
		@Override
		public Class<ReporterListExpression> getElementClass()
		{
			return ReporterListExpression.class;
		}
		@Override
		public void joinIfNeeded(DynamicReporterQuery query, ReporterListExpression element)
		{
			JoinType jt = element.getAutoJoinType();
			query.joinReporterList(element.getReporterList(), jt == null ? joinType : jt);
		}
		@Override
		public void reset(DynamicReporterQuery query)
		{}
		// -------------------------------------------
	}

	/**
		Auto joiner for master reporter table.
		@since 3.1
	*/
	static class MasterReporterJoiner
		implements AutoJoiner<DynamicReporterQuery, ReporterPropertyExpression>
	{
		private final JoinType joinType;
	
		private MasterReporterJoiner(JoinType joinType)
		{
			this.joinType = joinType;
		}
		
		/*
			From the AutoJoiner interface
			-------------------------------------------
		*/
		@Override
		public Class<DynamicReporterQuery> getQueryClass()
		{
			return DynamicReporterQuery.class;
		}
		@Override
		public Class<ReporterPropertyExpression> getElementClass()
		{
			return ReporterPropertyExpression.class;
		}
		@Override
		public void joinIfNeeded(DynamicReporterQuery query, ReporterPropertyExpression element)
		{
			query.joinMasterReporters(joinType);
		}
		@Override
		public void reset(DynamicReporterQuery query)
		{}
		// -------------------------------------------
	}

}
