/*
	$Id: Type.java 6539 2014-09-26 10:44:24Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.ParameterValueData;
import net.sf.basedb.core.data.IntegerParameterValueData;
import net.sf.basedb.core.data.LongParameterValueData;
import net.sf.basedb.core.data.FloatParameterValueData;
import net.sf.basedb.core.data.DoubleParameterValueData;
import net.sf.basedb.core.data.TextParameterValueData;
import net.sf.basedb.core.data.DateParameterValueData;
import net.sf.basedb.core.data.StringParameterValueData;
import net.sf.basedb.core.data.BooleanParameterValueData;
import net.sf.basedb.core.data.TimestampParameterValueData;
import net.sf.basedb.core.hibernate.TypeWrapper;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.text.NumberFormat;
import java.text.ParsePosition;
import java.util.Date;
import java.util.Map;
import java.util.HashMap;
import java.util.List;

/**
	This class defines constants for supported value types in BASE.
	@author Samuel, Nicklas
	@version 2.0
	@base.modified $Date: 2014-09-26 12:44:24 +0200 (fr, 26 sep 2014) $
*/
public enum Type
{

	/**
		Integer type.
	*/
	INT(TypeWrapper.INTEGER, 1, Integer.class, "int", "Integer", "IntegerValues", true, true, false)
	{
		@Override
		ParameterValueData<Integer> newParameterValueData()
		{
			return new IntegerParameterValueData();
		}
		@Override
		public int sizeOf(Object value)
		{
			return value == null ? 0 : 4;
		}
		@Override
		public Object parseString(String value)
			throws InvalidDataException
		{
			if (value == null) return null;
			return Integer.valueOf(Double.valueOf(value).intValue());
		}
		@Override
		public Number convertNumber(Number num)
		{
			return Integer.valueOf(num.intValue());
		}

		@Override
		public void writeValue(Serializable value, ObjectOutputStream out) 
			throws IOException
		{
			out.writeInt((Integer)value);
		}
		
		@Override
		public Serializable readValue(ObjectInputStream in) 
			throws IOException
		{
			return in.readInt();
		}
	},

	/**
		Long integer type.
	*/
	LONG(TypeWrapper.LONG, 2, Long.class, "long", "Long", "LongValues", true, true, false)
	{
		@Override
		ParameterValueData<Long> newParameterValueData()
		{
			return new LongParameterValueData();
		}
		@Override
		public int sizeOf(Object value)
		{
			return value == null ? 0 : 8;
		}
		@Override
		public Object parseString(String value)
			throws InvalidDataException
		{
			if (value == null) return null;
			return Long.valueOf(Double.valueOf(value).longValue());
		}
		@Override
		public Number convertNumber(Number num)
		{
			return Long.valueOf(num.longValue());
		}

		@Override
		public void writeValue(Serializable value, ObjectOutputStream out) 
			throws IOException
		{
			out.writeLong((Long)value);
		}
		
		@Override
		public Serializable readValue(ObjectInputStream in) 
			throws IOException
		{
			return in.readLong();
		}
	},

	/**
		Float type.
	*/
	FLOAT(TypeWrapper.FLOAT, 3, Float.class, "float", "Float", "FloatValues", true, true, false)
	{
		@Override
		ParameterValueData<Float> newParameterValueData()
		{
			return new FloatParameterValueData();
		}
		@Override
		public int sizeOf(Object value)
		{
			return value == null ? 0 : 4;
		}
		@Override
		public Object parseString(String value)
			throws InvalidDataException
		{
			if (value == null) return null;
			Float f = Float.valueOf(value);
			if (f.isNaN() || f.isInfinite()) f = null;
			return f;
		}
		@Override
		public Number convertNumber(Number num)
		{
			return Float.valueOf(num.floatValue());
		}
		
		@Override
		public void writeValue(Serializable value, ObjectOutputStream out) 
			throws IOException
		{
			out.writeFloat((Float)value);
		}
		
		@Override
		public Serializable readValue(ObjectInputStream in) 
			throws IOException
		{
			return in.readFloat();
		}
	},

	/**
		Double type.
	*/
	DOUBLE(TypeWrapper.DOUBLE, 4, Double.class, "double", "Double", "DoubleValues", true, true, false)
	{
		@Override
		ParameterValueData<Double> newParameterValueData()
		{
			return new DoubleParameterValueData();
		}
		@Override
		public int sizeOf(Object value)
		{
			return value == null ? 0 : 8;
		}
		@Override
		public Object parseString(String value)
			throws InvalidDataException
		{
			if (value == null) return null;
			Double d = Double.valueOf(value);
			if (d.isNaN() || d.isInfinite()) d = null;
			return d;
		}
		@Override
		public Number convertNumber(Number num)
		{
			return Double.valueOf(num.doubleValue());
		}
		
		@Override
		public void writeValue(Serializable value, ObjectOutputStream out) 
			throws IOException
		{
			out.writeDouble((Double)value);
		}
		
		@Override
		public Serializable readValue(ObjectInputStream in) 
			throws IOException
		{
			return in.readDouble();
		}
	},

	/**
		String type where maximum length allowed is 255 characters.
		@see #TEXT
	*/
	STRING(TypeWrapper.STRING, 5, String.class, "string", "String", "StringValues", true, false, false)
	{
		@Override
		ParameterValueData<String> newParameterValueData()
		{
			return new StringParameterValueData();
		}
		@Override
		public int sizeOf(Object value)
		{
			return value == null ? 0 : value.toString().length();
		}
		@Override
		public Object parseString(String value) 
			throws InvalidDataException
		{
			return value;
		}

		@Override
		public void writeValue(Serializable value, ObjectOutputStream out) 
			throws IOException
		{
			out.writeObject((String)value);
		}
		
		@Override
		public Serializable readValue(ObjectInputStream in) 
			throws IOException, ClassNotFoundException
		{
			return (String)in.readObject();
		}
	},

	/**
		Text blob type.
		@see #STRING
	*/
	TEXT(TypeWrapper.TEXT, 6, String.class, "text", "Text", "TextValues", false, false, false)
	{
		@Override
		ParameterValueData<String> newParameterValueData()
		{
			return new TextParameterValueData();
		}
		@Override
		public int sizeOf(Object value)
		{
			return value == null ? 0 : value.toString().length();
		}
		@Override
		public Object parseString(String value) 
			throws InvalidDataException
		{
			return value;
		}
		
		@Override
		public void writeValue(Serializable value, ObjectOutputStream out) 
			throws IOException
		{
			out.writeObject((String)value);
		}
		
		@Override
		public Serializable readValue(ObjectInputStream in) 
			throws IOException, ClassNotFoundException
		{
			return (String)in.readObject();
		}
	},

	/**
		Boolean type.
	*/
	BOOLEAN(TypeWrapper.BOOLEAN, 7, Boolean.class, "boolean", "Boolean", "BooleanValues", false, false, false)
	{
		@Override
		ParameterValueData<Boolean> newParameterValueData()
		{
			return new BooleanParameterValueData();
		}
		@Override
		public int sizeOf(Object value)
		{
			return value == null ? 0 : 1;
		}
		/**
			Updated in 2.7.1. Previously, only the string "true" was considered 
			a <code>true</code> value. Since 2.7.1, the empty string, "no", "0" 
			and "false" (case-insensitive) are considered <code>false</code>. 
			All other values are <code>true</code>, except null which returns null.
			@see net.sf.basedb.util.Values#getBoolean(String)
			@since 2.7.1
		*/
		@Override
		public Object parseString(String value)
			throws InvalidDataException
		{
			if (value == null) return null;
			return net.sf.basedb.util.Values.getBoolean(value);
		}
		
		@Override
		public void writeValue(Serializable value, ObjectOutputStream out) 
			throws IOException
		{
			out.writeBoolean((Boolean)value);
		}
		
		@Override
		public Serializable readValue(ObjectInputStream in) 
			throws IOException
		{
			return (Boolean)in.readBoolean();
		}
	},

	/**
		The parameter is a date value.
	*/
	DATE(TypeWrapper.DATE, 8, Date.class, "date", "Date", "DateValues", true, false, true)
	{
		@Override
		ParameterValueData<Date> newParameterValueData()
		{
			return new DateParameterValueData();
		}
		@Override
		public Object parseString(String value)
			throws InvalidDataException
		{
			return DateUtil.parseString(value);
		}
		@Override
		public int sizeOf(Object value)
		{
			return value == null ? 0 : 4;
		}
		
		@Override
		public void writeValue(Serializable value, ObjectOutputStream out) 
			throws IOException
		{
			out.writeLong(((Date)value).getTime());
		}
		
		@Override
		public Serializable readValue(ObjectInputStream in) 
			throws IOException
		{
			return new Date(in.readLong());
		}
	},

	/**
		The parameter is a date+time value.
		@since 2.16
	*/
	TIMESTAMP(TypeWrapper.TIMESTAMP, 9, Date.class, "timestamp", "Timestamp", "TimestampValues", false, false, true)
	{
		@Override
		ParameterValueData<Date> newParameterValueData()
		{
			return new TimestampParameterValueData();
		}
		@Override
		public Object parseString(String value)
			throws InvalidDataException
		{
			return DateUtil.parseTimestamp(value);
		}
		@Override
		public int sizeOf(Object value)
		{
			return value == null ? 0 : 8;
		}
		
		@Override
		public void writeValue(Serializable value, ObjectOutputStream out) 
			throws IOException
		{
			out.writeLong(((Date)value).getTime());
		}
		
		@Override
		public Serializable readValue(ObjectInputStream in) 
			throws IOException
		{
			return new Date(in.readLong());
		}
	};

	
	private static final Map<Integer, Type> valueMapping = new HashMap<Integer, Type>();
	private static final Map<String, Type> stringMapping = new HashMap<String, Type>();
	static
	{
		for(Type t : Type.values())
		{
			Type tt = valueMapping.put(t.getValue(), t);
			assert tt == null : "Another type with the value "+t.getValue()+" already exists";
			tt = stringMapping.put(t.getStringValue(), t);
			assert tt == null : "Another type with the string value "+t.getStringValue()+" already exists";
		}
	}
	


	private final TypeWrapper typeWrapper;
	private final int value;
	private final Class<?> valueClass;
	private final String tableName;
	private final String stringValue;
	private final String displayValue;
	private final boolean canEnumerate;
	private final boolean isNumerical;
	private final boolean isTemporal;

	private Type(TypeWrapper typeWrapper, int value, Class<?> valueClass, 
		String stringValue, String displayValue, String tableName, boolean canEnumerate, boolean isNumerical, boolean isTemporal)
	{
		this.typeWrapper = typeWrapper;
		this.value = value;
		this.valueClass = valueClass;
		this.stringValue = stringValue;
		this.displayValue = displayValue;
		this.tableName = tableName;
		this.canEnumerate = canEnumerate;
		this.isNumerical = isNumerical;
		this.isTemporal = isTemporal;
	}

	@Override
	public String toString()
	{
		return displayValue;
	}
	
	/**
		Get a wrapper for accessing underlying type information from Hibernate.
		@since 2.16
	*/
	public TypeWrapper getTypeWrapper()
	{
		return typeWrapper;
	}
	
	/**
		Get the integer value that is used when storing a type in the database.
	 	@return an int. 
	*/
	public int getValue()
	{
		return value;
	}

	/**
		Get the class of objects that can be used as values for this type.
	 	@return a Class  
		@see #isCorrectType(Object)
	*/
	public Class<?> getValueClass()
	{
		return valueClass;
	}
	
	/**
		Get the string value representation of this type.
	 	@return a java.lang.String 
	*/
	public String getStringValue()
	{
		return stringValue;
	}
	
	/**
		Get the type of this parameter as specified by the
		<code>java.sql.Types</code> class.
		@return The type-code of the parameter
	*/
	public int getSQLType()
	{
		return typeWrapper.getSQLType();
	}

	/**
		Get the name of the database table, parameter and annotation
		values of this value type are stored in.
		@since 2.9
	*/
	public String getTableName()
	{
		return tableName;
	}
	
	/**
		If it makes sense to use values of this type as an enumeration
		to choose from. For example, the <code>BOOLEAN</code> and <code>TEXT</code>
		type can't be enumerated.
	 	@return TRUE if it does, FALSE otherwise. 
	*/
	public boolean canEnumerate()
	{
		return canEnumerate;
	}
	
	/**
		If the values of this type are numerical values.
		@return TRUE if the values are numerical, FALSE otherwise
	*/
	public boolean isNumerical()
	{
		return isNumerical;
	}

	/**
		If the values of this type are date/time related.
		@return TRUE if the values are temporal, FALSE otherwise
		@since 2.16
	*/
	public boolean isTemporal()
	{
		return isTemporal;
	}

	
	/**
		Check if an object is a value of the correct type. Null is allowed.
	 	@param value Value object to be checked. 
		@throws InvalidDataException If the value isn't of the correct type
	*/
	public void validate(Object value)
		throws InvalidDataException
	{
		if (value != null && !isCorrectType(value))
		{
			throw new InvalidDataException("Value '"+value+"' is a '"+value.getClass()+"', not a '"+this+"'");
		}
	}

	/**
		Check if a list of values contain only objects of the correct type.
	 	@param values List of objects to be checked. 
		@throws InvalidDataException If the list contains one or more nulls or 
			values of incorrect object type
	*/
	public void validate(List<?> values)
		throws InvalidDataException
	{
		if (values == null || values.size() == 0) return;
		int i = 0;
		for (Object value : values)
		{
			if (value == null) throw new InvalidUseOfNullException("values["+i+"]");
			if (!isCorrectType(value))
			{
				throw new InvalidDataException("Value '"+value+"' is a '"+value.getClass()+"', not a '"+this+"'");
			}
			i++;
		}
	}

	/**
		Check if the value is an object of the correct type.
	 	@param value Value object to be checked. 
	 	@return TRUE if it is of correct type, FALSE otherwise.
	*/
	public boolean isCorrectType(Object value)
	{
		return valueClass.isInstance(value);
	}

	/**
		Create a new {@link ParameterValueData} subclass which is used to
		store values of this value type in the database.
	*/
	abstract ParameterValueData<?> newParameterValueData();

	/**
		Get the size in bytes an object of this type requires when
		stored in the database. If the value is null, 0 is returned, 
		otherwise the actual value is only used if it
		affects the size, ie. it is a string.
		@param value The value to get the size of
	 	@return an integer
	*/
	public abstract int sizeOf(Object value);
	
	/**
		Write a value of this type to the output stream. Eg. depending
		on which Type is ised one of ObjectOutputStream.write*(value)
		methods should be used. 
		@param value A value which must be of the correct type (not checked)
		@param out Stream to write to
		@since 3.4
	*/
	public abstract void writeValue(Serializable value, ObjectOutputStream out)
		throws IOException;
	
	/**
		Read a value of this type to the input stream. Eg. depending
		on which Type is ised one of ObjectInputStream.read*(value)
		methods should be used. 
		@param in Stream to read from
		@since 3.4
	*/
	public abstract Serializable readValue(ObjectInputStream in)
		throws IOException, ClassNotFoundException;
	
	/**
	 	Convert a number to the appropriate class. The default implementation
	 	returns null. Each numeric type overrides this method to create a number of
	 	the correct type.
	 	@param num The number to convert. 
	 	@return Null if this method not have been overridden.
	 	@since 2.2
	*/
	public Number convertNumber(Number num)
	{
		return null;
	}
	
	/**
		Parse a string and return a value of the correct type.
	 	@param value The string to get the type object from.
	  	@return a java.lang.Object or null if the parameter is null.
	 	@throws InvalidDataException If the argument is corrupt.
	*/
	public abstract Object parseString(String value)
		throws InvalidDataException;
	
	/**
		Parse a string and return a value of the correct type using a specific
		number format for numerical values, optionally ignoring unparsable numbers.
		@param value The string value to parse
		@param numberFormat The number format, or null to use {@link #parseString(String)}
		@param nullIfException TRUE to return null for unparsable numeric values, or
			FALSE to throw an exception
	 	@return a java.lang.Object, or null. 
		@throws NumberFormatException If the type is numerical and the string can't be 
			parsed as a number and the nullIfException parameter is false
		@since 2.4
	*/
	public Object parseString(String value, NumberFormat numberFormat, boolean nullIfException)
	{
		if (value == null)
		{
			return null;
		}
		else if (!isNumerical || numberFormat == null)
		{
			try
			{
				return parseString(value);
			}
			catch (RuntimeException t)
			{
				if (nullIfException) return null;
				throw t;
			}
		}
		else
		{
			value = value.trim();
			if ("NaN".equalsIgnoreCase(value)) return parseString("NaN");
			if (value.indexOf('e') > 0) value = value.replace('e', 'E');
			if (value.indexOf("E+") > 0) value = value.replace("E+", "E");
			ParsePosition pos = new ParsePosition(0);
			Number num = numberFormat.parse(value, pos);
			if (pos.getIndex() < value.length())
			{
				if (nullIfException)
				{
					return null;
				}
				else
				{
					throw new NumberFormatException("For input string: \"" + value + "\"");
				}
			}
			return convertNumber(num);
		}
	}
	
	/**
		Parse an array of strings and try to convert them to values
		of the correct object type.
	 	@param values An array of values to parse through. 
		@return An array of objects, or null of the array is null
		@throws InvalidDataException If at least one of the strings could not 
			be parsed to the correct type
	*/
	public Object[] parseStrings(String[] values)
		throws InvalidDataException
	{
		if (values == null) return null;
		Object[] o = new Object[values.length];
		for (int i = 0; i < values.length; ++i)
		{
			o[i] = parseString(values[i]);
		}
		return o;
	}

	/**
		Get the <code>Type</code> object when you know the integer code.
	 	@param value The integer code to get the Type object for. 
	 	@return a net.sf.basedb.core.Type object
	*/
	public static Type fromValue(int value)
	{
		Type t = valueMapping.get(value);
		return t;
	}

	/**
		Get the <code>Type</code> object when you know the string value.
	 	@param value The string value to get the Type object for. 
	 	@return a net.sf.basedb.core.Type object.
	*/
	public static Type fromValue(String value)
	{
		Type t = stringMapping.get(value);
		return t;
	}
	
	/**
		Try to find a matching Type instance given a Hibernate Type object.
		@param hibernateType The Hibernate type
		@return A Type instance or null if not found
		@since 2.15
	*/
	public static Type fromHibernateType(org.hibernate.type.Type hibernateType)
	{
		for (Type t : Type.values())
		{
			if (t.getTypeWrapper().getHibernateType().equals(hibernateType)) return t;
		}
		// TODO -- The Hibernate type model is more fine-grained 
		// We can maybe check some equivalent types here
		return null;
	}

}
