/**
	$Id $

	Copyright (C) 2013 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.log;

import java.lang.reflect.Method;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.WeakHashMap;

import net.sf.basedb.core.data.NotLoggable;
import net.sf.basedb.core.data.PropertyPathProtected;

/**
	This class collect some static information related to logging for
	entities. Currently, it will scan all entity classes (lazily) for
	methods annotated with {@link NotLoggable} so that the logging 
	implementation quickly can check if changes to a given property should 
	be ignored or not.

	@author Nicklas
	@since 3.3
	@base.modified $Date$
*/
public class EntityLogOptions
{

	private static final Map<Class<?>, EntityLogOptions> options = new WeakHashMap<Class<?>, EntityLogOptions>();

	/**
		Get the logging options for the given entity.
		@return Logging options or null if there are no options
	*/
	public static EntityLogOptions getEntityLogOptions(Object entity)
	{
		if (entity == null) return null;
		Class<?> entityClass = entity.getClass();
		return getEntityLogOptionsForClass(entityClass);
	}

	/**
		Get the logging options for the given class of entities.
		@return Logging options or null if there are no options
	*/
	public static EntityLogOptions getEntityLogOptionsForClass(Class<?> entityClass)
	{
		if (entityClass == null) return null;
		
		EntityLogOptions logOptions = null;
		if (options.containsKey(entityClass))
		{
			logOptions = options.get(entityClass);
		}
		else
		{
			logOptions = create(entityClass);
			options.put(entityClass, logOptions);
		}
		return logOptions;
	}
	
	private static EntityLogOptions create(Class<?> entityClass)
	{
		EntityLogOptions logOptions = new EntityLogOptions(entityClass);
		// No need to keep object around if it doesn't contain any info
		return logOptions.size == 0 ? null : logOptions;
	}

	
	private final Set<String> notLoggableProperties;
	private final Set<String> propertyPathProtectedProperties;
	private final int size;
	
	private EntityLogOptions(Class entityClass)
	{
		notLoggableProperties = new HashSet<String>();
		propertyPathProtectedProperties = new HashSet<String>();
		
		// Check methods declared on this class for the @NotLoggable annotation
		for (Method m : entityClass.getDeclaredMethods())
		{
			String propertyName = getPropertyName(m.getName());
			if (m.isAnnotationPresent(NotLoggable.class)) 
			{
				notLoggableProperties.add(propertyName);
			}
			if (m.isAnnotationPresent(PropertyPathProtected.class))
			{
				propertyPathProtectedProperties.add(propertyName);
			}
		}
		
		// Merge superclass options
		merge(EntityLogOptions.getEntityLogOptionsForClass(entityClass.getSuperclass()));
		
		// and all implemented interfaces
		for (Class<?> ifc : entityClass.getInterfaces())
		{
			merge(EntityLogOptions.getEntityLogOptionsForClass(ifc));
		}
		
		size = notLoggableProperties.size() + propertyPathProtectedProperties.size();
	}
	
	/**
		Merge log options from the other object into this
		object. Merging is usually performed due to the other
		options object representing the log options for a
		superclass or implemented interface.
	*/
	private void merge(EntityLogOptions other)
	{
		if (other == null) return;	
		// Copy all @NotLoggable properties
		notLoggableProperties.addAll(other.notLoggableProperties);
		// Copy all @PropertyPathProtected propertie
		propertyPathProtectedProperties.addAll(other.propertyPathProtectedProperties);
	}
	
	/**
		Convert a method name into a property name. Typically
		remove 'get', 'set' or 'is' prefix and then lowercase the first
		letter.
	*/
	private String getPropertyName(String methodName)
	{
		String propertyName = methodName;
		if (methodName.startsWith("get") || methodName.startsWith("set"))
		{
			propertyName = methodName.substring(3);
		}
		else if (methodName.startsWith("is"))
		{
			propertyName = methodName.substring(2);
		}
		return propertyName.length() == 0 ? "" : propertyName.substring(0, 1).toLowerCase() + propertyName.substring(1);
	}

	/**
		Check if the method corresponding to the given property name
		has been annotated with the {@link NotLoggable} annotation or 
		not.
		@param propertyName The name of the property, not including 'get', 'set' or 'is' prefix
		@return TRUE if the {@link NotLoggable} annotation is present, FALSE otherwise
	*/
	public boolean isNotLoggable(String propertyName)
	{
		return notLoggableProperties.contains(propertyName);
	}
	
	/**
		Check if the method corresponding to the given property name
		has been annotated with the {@link PropertyPathProtected} annotation or 
		not.
		@param propertyName The name of the property, not including 'get', 'set' or 'is' prefix
		@return TRUE if the {@link PropertyPathProtected} annotation is present, FALSE otherwise
	*/
	public boolean isPropertyPathProtected(String propertyName)
	{
		return propertyPathProtectedProperties.contains(propertyName);
	}
}
