/**
	$Id: AnnotationSnapshot.java 6721 2015-02-10 12:26:32Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.snapshot;

import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;

import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.Annotation;
import net.sf.basedb.core.AnnotationSet;
import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.Unit;
import net.sf.basedb.core.data.AnnotationData;
import net.sf.basedb.core.data.AnnotationSetData;
import net.sf.basedb.util.filter.Filter;
import net.sf.basedb.util.units.UnitConverter;

/**
	Stores information about a single primary or inherited 
	annotation or an inherited annotation set. Actual 
	annotation values are only stored for primary annotations.
	<p>
	<b>IMPORTANT NOTE TO DEVELOPERS!!</b>
	<p>
	Do not forget to synchronize all serialization code and to increment
	the {@link AnnotationSetSnapshot#FILE_VERSION} in case the number of variables 
	that needs serialization in this class or in the {@link AnnotationSetSnapshot} 
	changes. 

	@author Nicklas
	@version 2.14
	@base.modified $Date: 2015-02-10 13:26:32 +0100 (ti, 10 feb 2015) $
*/
public class AnnotationSnapshot
	implements Serializable
{

	private static final long serialVersionUID = 5826828832517594630L;
	
	private transient boolean inherited;
	private transient int annotationId;
	private transient int annotationSetId;
	private transient int annotationTypeId;
	private transient int unitId;
	private transient Type valueType;
	private transient List<? extends Serializable> values;
	private transient int itemId;
	private transient Item itemType;
	
	/**
		Create a new, empty annotation snapshot.
	*/
	public AnnotationSnapshot()
	{}
	
	/**
		Initialize the snapshot with information from the given
		primary annotation.
		@param primary A primary annotation
	*/
	@SuppressWarnings("unchecked")
	void initPrimary(AnnotationData primary)
	{
		this.inherited = false;
		this.annotationId = primary.getId();
		this.annotationSetId = primary.getAnnotationSet().getId();
		this.annotationTypeId = primary.getAnnotationType().getId();
		this.unitId = primary.getUnit() == null ? 0 : primary.getUnit().getId();
		this.valueType = Type.fromValue(primary.getAnnotationType().getValueType());
		this.values = new ArrayList<Serializable>((List<Serializable>)primary.getValues().getValues());
	}
	
	/**
		Initialize the snapshot with information from the given
		inherited annotation.
		@param inherited An inherited annotation
	*/
	void initInherited(AnnotationData inherited)
	{
		this.inherited = true;
		this.annotationId = inherited.getId();
		this.annotationSetId = inherited.getAnnotationSet().getId();
		this.annotationTypeId = inherited.getAnnotationType().getId();
	}
	
	/**
		Initialize the snapshot with information from the given
		inherited annotation set.
		@param inherited An inherited annotation set
	*/
	void initInherited(AnnotationSetData inherited)
	{
		this.inherited = true;
		this.annotationSetId = inherited.getId();
	}
	
	/**
		Initializes this snapshot as an inherited annotation
		with complete data from another primary annotation
		snapshot.
		@param primary A snapshot of a primary annotation
	*/
	void initInherited(AnnotationSnapshot primary)
	{
		this.inherited = true;
		this.annotationId = primary.annotationId;
		this.annotationTypeId = primary.annotationTypeId;
		this.annotationSetId = primary.annotationSetId;
		this.unitId = primary.unitId;
		this.valueType = primary.valueType;
		this.values = primary.values;
		this.itemId = primary.itemId;
		this.itemType = primary.itemType;
	}
	
	/**
		Is this annotation a primary or inherited annotation?
	*/
	public boolean isInherited()
	{
		return inherited;
	}
	
	/**
		Get the id of the annotation this is a snapshot of.
		@return The id or 0 if the snapshot is from an inherited 
			annotation set
	*/
	public int getAnnotationId()
	{
		return annotationId;
	}
	
	/**
		Get the annotation this is a snapshot of.
		@return An Annotation object or null if this snapshot is from
			an inherited annotation set
	*/
	public Annotation getAnnotation(DbControl dc)
	{
		return annotationId == 0 ? null : Annotation.getById(dc, annotationId);
	}
	
	/**
		Check if the logged in user has the request permission on this annotation
		or annotation set.
		@param dc The DbControl to use for database access
		@param permission The requested permission
		@return TRUE if the logged in user has the requested permission, FALSE otherwise
		@since 3.1.1
	*/
	public boolean hasPermission(DbControl dc, Permission permission)
	{
		if (annotationTypeId != 0 && permission == Permission.READ)
		{
			// Special case to avoid loading each Annotation and it's associated item from
			// the database just for the permission check which is only based on the permission
			// from the annotation type in any case
			// See tickets #1693 and #1685
			try
			{
				AnnotationType at = getAnnotationType(dc);
				return true;
			}
			catch (RuntimeException ex)
			{}
		}
		else if (annotationId != 0)
		{
			try
			{
				Annotation a = getAnnotation(dc);
				return a.hasPermission(permission);
			}
			catch (RuntimeException ex)
			{}
		}
		else if (annotationSetId != 0)
		{
			try
			{
				AnnotationSet a = getAnnotationSet(dc);
				return a.hasPermission(permission);
			}
			catch (RuntimeException ex)
			{}
		}
		return false;
	}
	
	/**
		Get the id of the annotation set this annotation belongs to.
	*/
	public int getAnnotationSetId()
	{
		return annotationSetId;
	}

	/**
		Get the annotation set this annotation belongs to.
	*/
	public AnnotationSet getAnnotationSet(DbControl dc)
	{
		return AnnotationSet.getById(dc, getAnnotationSetId());
	}
	
	/**
		Get the id of the annotation type.
		@return The annotation type id, or 0 if this snapshot is from
			an inherited annotation set
	*/
	public int getAnnotationTypeId()
	{
		return annotationTypeId;
	}

	/**
		Get the annotation type.
		@return An AnnotationType object or null if this snapshot is
			from an inherited annotation set
	*/
	public AnnotationType getAnnotationType(DbControl dc)
	{
		return annotationTypeId == 0 ? null :AnnotationType.getById(dc, getAnnotationTypeId());
	}

	/**
		Get the id of the unit that the annotation values should
		use when displayed. 
		@return The unit id, or 0 if this annotatation doesn't use units
		@since 3.4
	*/
	public int getUnitId()
	{
		return unitId;
	}
	
	/**
		Get the unit.
		@return A Unit object or null if this annotation doesn't use units
		@since 3.4
	*/
	public Unit getUnit(DbControl dc)
	{
		return unitId == 0 ? null : Unit.getById(dc, unitId);
	}

	/**
		Get the annotation values. NOTE! The values are only available
		for primary annotations and for inherited annotations returned by
		{@link SnapshotManager#findAnnotations(DbControl, AnnotationSetSnapshot, Filter, boolean)}
		
		@return A list with the values
	*/
	public List<? extends Serializable> getValues()
	{
		return values;
	}
	
	/**
		Get the annotations values, optionally converted to some specific unit
		and type of values.
		
		NOTE! To create a unit converter that converts that values to the unit
		specified by {@link #getUnit(DbControl)}:
		
		this.getUnit(dc).getUnitConverter(this.getAnnotationType(dc).getDefaultUnit())
		
		@param converter A converter for converting (numerical) values,
			or null to not convert the values
		@param valueType The return type of the converted values (ignored if
			no converter is specified)
		@return A list with the values
	*/
	public List<? extends Serializable> getValues(UnitConverter converter, Type valueType)
	{
		if (converter == null) return values;
		List<Serializable> convertedValues = new ArrayList<Serializable>(values.size());
		for (Serializable value : values)
		{
			if (value instanceof Number)
			{
				value = valueType.convertNumber(converter.convertToSpecificUnit(((Number)value).doubleValue()));
			}
			convertedValues.add(value);
		}
		return convertedValues;
	}
	
	/**
	 	Get the id of the item this annotation belongs to. NOTE! this value
	 	is only available on annotations returned from the call to:
		{@link SnapshotManager#findAnnotations(DbControl, AnnotationSetSnapshot, Filter, boolean)}
	*/
	public int getItemId()
	{
		return itemId;
	}
	
	/**
		Get the item that this annotation belongs to.
		@return The item, or null if the item is not know
	*/
	public Annotatable getItem(DbControl dc)
	{
		return itemId == 0 ? null : (Annotatable)getItemType().getById(dc, getItemId());
	}

	/**
		Get the item type of the item this annotation belongs to. 
		NOTE! this value is only available on annotations returned from the call to:
		{@link SnapshotManager#findAnnotations(DbControl, AnnotationSetSnapshot, Filter, boolean)}

		@return The item type, or null if not known
	*/
	public Item getItemType()
	{
		return itemType;
	}

	void setItem(int itemId, Item itemType)
	{
		this.itemId = itemId;
		this.itemType = itemType;
	}

	void serializeObject(java.io.ObjectOutputStream out)
		throws IOException
	{
		writeObject(out);
	}
	
	private void writeObject(java.io.ObjectOutputStream out)
		throws IOException
	{
		out.writeBoolean(inherited);
		out.writeInt(annotationId);
		out.writeInt(annotationSetId);
		out.writeInt(annotationTypeId);
		out.writeInt(unitId);
		out.writeInt(valueType == null ? 0 : valueType.getValue());
		out.writeInt(values == null ? 0 : values.size());
		if (values != null)
		{
			for (Serializable tmp : values)
			{
				valueType.writeValue(tmp, out);
			}
		}
	}
	
	void deserializeObject(java.io.ObjectInputStream in)
		throws IOException, ClassNotFoundException
	{
		readObject(in);
	}
	private void readObject(java.io.ObjectInputStream in)
		throws IOException, ClassNotFoundException
	{
		inherited = in.readBoolean();
		annotationId = in.readInt();
		annotationSetId = in.readInt();
		annotationTypeId = in.readInt();
		unitId = in.readInt();
		valueType = Type.fromValue(in.readInt());
		int numValues = in.readInt();
		if (numValues > 0)
		{
			List<Serializable> values = new ArrayList<Serializable>(numValues);
			for (int i = 0; i < numValues; ++i)
			{
				values.add(valueType.readValue(in));
			}
			this.values = values;
		}
	}
	
}
