/**
	$Id: DataFileValidator.java 6089 2012-08-22 08:09:59Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.validator;

import java.util.Collections;
import java.util.List;

import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileSet;
import net.sf.basedb.core.FileSetMember;
import net.sf.basedb.core.FileStoreEnabled;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.ItemSubtypeFileType;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.Platform;
import net.sf.basedb.core.PlatformFileType;
import net.sf.basedb.core.PlatformVariant;
import net.sf.basedb.core.Subtypable;
import net.sf.basedb.core.UsableDataFileType;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.overview.Fix;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.Validator;
import net.sf.basedb.util.overview.Node;

/**
	Validator implementation for data files. The node containing
	the data files is expected to live in a folder-type node that
	has a {@link FileStoreEnabled} item as it's parent (if there 
	are more than one file) or as an immediate child node to
	it's parent item (if there is only one file).
	Validation rules:
	<ul>
	<li>Denied access to file: {@link Validator#DENIED_FILE}
	<li>Denied access to file type: {@link Validator#DENIED_DATAFILETYPE}
	<li>File is not a valid file: {@link Validator#DATAFILE_INVALID}
	<li>File has not been validated: {@link Validator#DATAFILE_NOTVALIDATED}
	<li>File type doesn't match parent item type: {@link Validator#DATAFILE_INVALID_ITEM}
	<li>File type is not part of used platform: {@link Validator#DATAFILE_NOTPLATFORM}
	<li>Too many files of a single-file file type: {@link Validator#DATAFILE_TOOMANYFILES}
	<li>Missing a required file: {@link Validator#MISSING_DATAFILE}
	</ul>
	
	@author Nicklas
	@version 2.10
	@base.modified $Date: 2012-08-22 10:09:59 +0200 (on, 22 aug 2012) $
*/
public class DataFileValidator
	extends BasicNodeValidator<FileSetMember>
{
	
	public DataFileValidator()
	{
		super(null, null);
	}
	
	/*
		From BasicNodeValidator class
		-----------------------------
	*/
	@Override
	public void postValidate(DbControl dc, OverviewContext context, Node node, Node parentNode)
	{
		super.postValidate(dc, context, node, parentNode);
		FileSetMember member = (FileSetMember)node.getItem(dc);
		// The parent node is either a folder (if we have more than one file)
		// or the actual parent (for a single file)
		FileStoreEnabled parentItem = (FileStoreEnabled)(parentNode.getNodeType() == Node.Type.FOLDER ? 
				parentNode.getParent().getItem(dc) : parentNode.getItem(dc));
		
		// Check that we have access to the file and it's file type
		DataFileType fileType = null;
		File file = null;
		try
		{
			fileType = member.getDataFileType();
		}
		catch (PermissionDeniedException ex)
		{}
		try
		{
			file = member.getFile();
		}
		catch (PermissionDeniedException ex)
		{}
		if (fileType == null)
		{
			context.createFailure(Validator.DENIED_DATAFILETYPE, node, null);
		}
		if (file == null)
		{
			context.createFailure(Validator.DENIED_FILE, node, null);
		}
		
		// Check if a validator has determined that the file IS invalid
		if (Boolean.FALSE.equals(member.isValid()))
		{
			String errorMsg = fileType == null ? "Data file " : fileType.getName();
			errorMsg += " is not valid: " + member.getErrorMessage();
			context.createFailure(Validator.DATAFILE_INVALID, node, errorMsg,
				new Fix("Select another file", (BasicItem)parentItem, member)
			);
		}
		
		if (fileType != null)
		{
			// Check if the file hasn't been validated even if a validator exists
			if (fileType.hasActiveValidator(dc) && member.isValid() == null)
			{
				context.createFailure(Validator.DATAFILE_NOTVALIDATED, node, 
					fileType.getName() + " has not been validated", 
					new Fix("Validate the file / Select another file", (BasicItem)parentItem, member));
			}

			// Check that the file type is matching the parent item's type
			if (fileType.getItemType() != parentItem.getType())
			{
				context.createFailure(Validator.DATAFILE_INVALID_ITEM, node, 
					"Data file '" + fileType.getName() + "' can't be used on item type: " + parentItem.getType(), 
					new Fix("Remove file", (BasicItem)parentItem, member)
				);
			}

			// Check that the file type is part of the used platform/variant/subtype
			try
			{
				Platform platform = parentItem.getPlatform();
				PlatformVariant variant = parentItem.getVariant();
				if (platform != null)
				{
					// The item to apply fixes on, variant or platform
					String pvName = null;
					BasicItem pvItem = null;
					if (variant != null)
					{
						pvItem = variant;
						pvName = variant.getName();
					}
					else
					{
						pvItem = platform;
						pvName = platform.getName();
					}
					if (platform.getFileType(fileType, variant, false) == null)
					{
						context.createFailure(Validator.DATAFILE_NOTPLATFORM, node, 
								fileType.getName() + " can't be used with '" + pvName + "' platform", 
								new Fix("Remove '" + fileType.getName() + "' from " + parentItem.getType().toString().toLowerCase(), 
									(BasicItem)parentItem, fileType),
								new Fix("Add '" + fileType.getName() +"' to '" + pvName + "' platform", pvItem, fileType)
							);
					}
				}
				else if (parentItem instanceof Subtypable)
				{
					ItemSubtype subtype = ((Subtypable)parentItem).getItemSubtype();
					if (subtype != null)
					{
						if (subtype.getAssociatedDataFileType(fileType, false) == null)
						{
							context.createFailure(Validator.DATAFILE_NOTPLATFORM, node, 
									fileType.getName() + " can't be used with '" + subtype.getName() + "' subtype", 
									new Fix("Remove '" + fileType.getName() + "' from " + parentItem.getType().toString().toLowerCase(), 
										(BasicItem)parentItem, fileType),
									new Fix("Add '" + fileType.getName() +"' to '" + subtype.getName() + "' subtype", subtype, fileType)
								);
						}
					}
				}
			}
			catch (PermissionDeniedException ex)
			{}		
		}
	}
	
	/**
		@return Always "null"
	*/
	@Override
	protected Fix getMissingItemFix(DbControl dc, Node parentNode)
	{
		return null;
	}

	@Override
	public void postValidateFolder(DbControl dc, OverviewContext context, 
			Node folderNode, Node parentNode)
	{
		super.postValidateFolder(dc, context, folderNode, parentNode);
		FileStoreEnabled parentItem = (FileStoreEnabled)parentNode.getItem(dc);
		
		// Check that we have linked a file for all file types that
		// are required by the platform and that the multiplicity setting
		// is ok
		try
		{
			Platform platform = parentItem.getPlatform();
			PlatformVariant variant = parentItem.getVariant();
			BasicItem pvItem = null;
			ItemSubtype subtype = parentItem instanceof Subtypable ? 
					((Subtypable)parentItem).getItemSubtype() : null;
			// List for holding all file types defined by the platform or subtype
			List<? extends UsableDataFileType> fileTypes = Collections.emptyList();
			if (platform != null)
			{
				pvItem = variant != null ? variant : platform;
				// Create a query that returns all required platform file types
				ItemQuery<PlatformFileType> query = platform.getFileTypes(variant, variant == null);
				query.join(Hql.innerJoin("dataFileType", "dft"));
				query.restrict(
					Restrictions.eq(
						Hql.property("dft", "itemType"), 
						Expressions.integer(parentNode.getItemType().getValue())
					)
				);
				fileTypes = query.list(dc);
			}
			if (subtype != null)
			{
				pvItem = subtype;
				// Create a query that returns all required subtype file types
				ItemQuery<ItemSubtypeFileType> query = subtype.getDataFileTypes();
				query.join(Hql.innerJoin("dataFileType", "dft"));
				query.restrict(
					Restrictions.eq(
						Hql.property("dft", "itemType"), 
						Expressions.integer(parentNode.getItemType().getValue())
					)
				);
				fileTypes = query.list(dc);
			}
			FileSet fs = parentItem.hasFileSet() ? parentItem.getFileSet() : null;
			for (UsableDataFileType fileType : fileTypes)
			{
				DataFileType dft = fileType.getDataFileType();
				int numMembers = fs == null ? 0 : fs.getNumMembers(dft);
				if (fileType.isRequired() && numMembers == 0)
				{
					context.createFailure(Validator.MISSING_DATAFILE, parentNode, 
							"Missing '" + dft.getName() + "'", 
							new Fix("Add '" + dft.getName() + "' to " + parentItem.getType().toString().toLowerCase(), 
									(BasicItem)parentItem, dft),
							new Fix("Remove 'Required' flag from '" + dft.getName() + "'", pvItem, dft)
						);
				}
				if (!fileType.getAllowMultiple() && numMembers > 1)
				{
					context.createFailure(Validator.DATAFILE_TOOMANYFILES, parentNode, 
							"Too many files: '" + dft.getName() + "'", 
							new Fix("Remove '" + dft.getName() + "' from " + parentItem.getType().toString().toLowerCase(), 
									(BasicItem)parentItem, dft),
							new Fix("Check 'Allow multiple files' on '" + dft.getName() + "'", pvItem, dft)
						);
				}
			}
		}
		catch (PermissionDeniedException ex)
		{}
	}
	// ----------------------------
		
}
