/**
	$Id: IlluminaRawDataImporter.java 6127 2012-09-14 07:28:35Z nicklas $

	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins;

import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.BioMaterialEventSource;
import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DerivedBioAssay;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.Extract;
import net.sf.basedb.core.FeatureIdentificationMethod;
import net.sf.basedb.core.File;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PhysicalBioAssay;
import net.sf.basedb.core.Platform;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.Protocol;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.RawDataBatcher;
import net.sf.basedb.core.RawDataProperty;
import net.sf.basedb.core.RawDataType;
import net.sf.basedb.core.RawDataTypes;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.Software;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.SystemItems;
import net.sf.basedb.core.Trashcan;
import net.sf.basedb.core.data.RawData;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.plugins.util.Parameters;
import net.sf.basedb.util.Coordinate;
import net.sf.basedb.util.error.SimpleErrorHandler;
import net.sf.basedb.util.fuzzy.StringMatcher;
import net.sf.basedb.util.parser.ColumnMapper;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.Mapper;
import net.sf.basedb.util.parser.FlatFileParser.Data;
import net.sf.basedb.util.parser.FlatFileParser.Line;
import net.sf.basedb.util.parser.FlatFileParser.LineType;

/**
	An importer plug-in for Illumina raw data. The plug-in will create one or more
	raw bioassays from the data in the file. Optionally, it will add the raw bioassays
	to an experiment.
	<p>
	Since the data files doesn't have any coordinate information the importer will 
	use one of the following methods.
	
	<ul>
	<li>
		If no array design has been selected or if the array design is using the
		{@link FeatureIdentificationMethod#COORDINATES} method for identifying features:
		The plug-in create fake coordinates like this: block=1, column=1, row=linenumber in file.
		The linenmbers start with 1 at the first data line, ie. header lines are not counted.

	<li>
		If the array design uses the {@link FeatureIdentificationMethod#POSITION}
		method for identiyfing features: The plug-in sets the position=line number in file.
		The linenmbers start with 1 at the first data line, ie. header lines are not counted.
		
	<li>
		If the array design uses the {@link FeatureIdentificationMethod#FEATURE_ID}
		method for identifying features. The plug-in assumes that the feature ID is the
		same as the reporter ID.
	</ul>
	
	<p>
	NOTE! Since the methods are not conflicting with each other, there will not
	be an actual check which method to use by this plug-in. We will simple set all
	values as specified above and let the BASE core handle the identification.

	@author nicklas
	@version 2.4
	@base.modified $Date: 2012-09-14 09:28:35 +0200 (fr, 14 sep 2012) $
*/
public class IlluminaRawDataImporter
	extends AbstractFlatFileImporter
	implements InteractivePlugin
{

	
	private static final Set<GuiContext> guiContexts = 
		Collections.unmodifiableSet(new HashSet<GuiContext>(
			Arrays.asList(
				new GuiContext(Item.RAWBIOASSAY, GuiContext.Type.LIST),
				new GuiContext(Item.EXPERIMENT, GuiContext.Type.ITEM)
			)
		));
	
	private static final PluginParameter<String> featureIdentificationParameter = 
		new PluginParameter<String>(
			"featureIdentification",
			"Identify features by",
			"Choose which method to use for identifying features. If no value " +
			"is selected the identification method used on the array design is used: \n\n" +
			"* COORDINATES: Use auto-generated coordinates (block = 1, column = 1, row = line number in file)\n" +
			"* POSITION: Use auto-generated position = line number in file\n" +
			"* FEATURE_ID: Use TargetID (requires that each Target ID only appears once in the file)\n\n" +
			"NOTE! This parameter has no meaning unless an array design is selected",
			new StringParameterType(255, null, false, 1, 0, 0, 
				Arrays.asList(new String[] {"COORDINATE", "POSITION", "FEATURE_ID" } ))
		);
	
	private static final PluginParameter<String> invalidColumnsErrorParameter = new PluginParameter<String>(
			"invalidColumnsError",
			"Mismatch of columns",
			"What to do if the file doesn't include the same columns for all arrays, for example " +
			"if one array has AVG_Signal and MIN_Signal but another only has AVG_Signal\n\n"+
			"ignore = Ignore this and import the data that is there\n"+
			"fail = Stop with an error message",
			new StringParameterType(255, null, false, 1, 0, 0, 
				Arrays.asList( new String[] { "ignore", "fail"} ))
		);

	private static final PluginParameter<String> missingReporterErrorParameter = new PluginParameter<String>(
			"missingReporterError",
			"Reporter not found",
			"How to handle errors that are caused by a reporter not beeing present in the database. " +
				"NOTE! This option is not used when an array design has been selected. " +
				"If not specified the default error handling is used. \n\n"+
				"null = Import the data but set the reporter to null\n"+
				"skip = Skip the current data line and continue\n"+
				"fail = Stop with an error message",
				new StringParameterType(255, null, false, 1, 0, 0, 
						Arrays.asList( new String[] { "null", "skip", "fail"} ))
			);

	private static final PluginParameter<String> featureMismatchErrorParameter = new PluginParameter<String>(
			"featureMismatchError",
			"Feature mismatch",
			"How to handle errors that are caused by the data not matching the feature on the array design. " +
				"This can for example be caused by invalid spot coordinates, Target ID or a different reporter. " +
				"NOTE! This parameter has no meaning unless an array design is selected. " +
				"If not specified the default error handling is used.\n\n"+
				"smart = SKIP lines that have no feature on the array design, " +
				"ACCEPT lines that has a feature with a NULL reporter, " +
				"FAIL lines that has a feature with different reporter\n" +
				"skip = Skip the current data line and continue\n"+
				"fail = Stop with an error message",
				new StringParameterType(255, "smart", false, 1, 0, 0, 
					Arrays.asList( new String[] { "smart", "skip", "fail"} ))
			);
	
	/**
		Section definition for grouping associations to other items:
		scan, software, protocol and experiment
	*/
	private static final PluginParameter<String> associationsSection = new PluginParameter<String>(
		"associationsSection",
		"Associations",
		"Select items that you want to associate the imported raw bioassays with", 
		null
	);

	private RequestInformation configureJob;
	
	public IlluminaRawDataImporter()
	{}
	
	/*
		From the Plugin interface
		-------------------------------------------
	*/
	@Override
	public boolean supportsConfigurations()
	{
		return false;
	}
	@Override
	public boolean requiresConfiguration()
	{
		return false;
	}
	// -------------------------------------------
	
	/*
		From the InteractivePlugin interface
		-------------------------------------------
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	/**
		If used from an experiment context, verify that the experiment
		is an 'illumina' experiment and that the logged in user has
		permission to write.
	*/
	@Override
	public String isInContext(GuiContext context, Object item)
	{
		String message = null;
		if (context.getItem() == Item.EXPERIMENT)
		{
			if (item == null)
			{
				message = "The object is null";
			}
			else if (!(item instanceof Experiment))
			{
				message = "The object is not an experiment: " + item;
			}
			else
			{
				Experiment experiment = (Experiment)item;
				RawDataType illumina = RawDataTypes.getSafeRawDataType("illumina");
				if (illumina != experiment.getRawDataType())
				{
					message = "The experiment can't contain Illumina data: " + item;
				}
				else
				{
					experiment.checkPermission(Permission.WRITE);
				}
			}
		}
		return message;
	}
	@Override
	public RequestInformation getRequestInformation(GuiContext context, String command) 
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestInformation = getConfigureJobParameters(context);
		}
		return requestInformation;
	}
	@Override
	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				RequestInformation ri = getConfigureJobParameters(context);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found in the request", errors);
					return;
				}
				
				storeValue(job, request, fileParameter);
				storeValue(job, request, ri.getParameter(Parameters.CHARSET_PARAMETER));
				storeValue(job, request, ri.getParameter(Parameters.DECIMAL_SEPARATOR_PARAMETER));
				
				// Associations
				storeValue(job, request, ri.getParameter("experiment"));
				storeValues(job, request, ri.getParameter("bioAssays"));
				storeValue(job, request, ri.getParameter("arrayDesign"));
				storeValue(job, request, featureIdentificationParameter);
				storeValue(job, request, ri.getParameter("protocol"));
				storeValue(job, request, ri.getParameter("software"));
				
				// Error handling parameters
				storeValue(job, request, defaultErrorParameter);
				storeValue(job, request, missingReporterErrorParameter);
				storeValue(job, request, featureMismatchErrorParameter);
				storeValue(job, request, invalidColumnsErrorParameter);
				storeValue(job, request, numberFormatErrorParameter);
				
				File file = (File)job.getValue("file");
				if (context.getItem() == Item.EXPERIMENT)
				{
					Experiment exp = (Experiment)job.getValue("experiment");
					response.setSuggestedJobName("Import Illumina raw bioassays to experiment '" + exp.getName() + 
							"' from file '" + file.getName() + "'");
				}
				else
				{
					response.setSuggestedJobName("Import Illumina raw bioassays from file '" + 
							file.getName() + "'");
				}
				
				response.setDone("Job configuration complete", Job.ExecutionTime.SHORT);
			}
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
	}
	// -------------------------------------------
	
	/*
		From the AbstractFlatFileReporter class
		-------------------------------------------
	*/
	private RawDataType illumina;
	private FlatFileParser ffp;
	private DbControl dc;
	private Experiment experiment;
	private List<DerivedBioAssay> bioAssays;
	private ArrayDesign design;
	private Software software;
	private Protocol protocol;
	private FeatureIdentificationMethod fiMethod;
	private List<RawBioAssay> rawBioAssays;
	private List<BatchAndMapHolder> holders;
	private List<Line> headerLines;
	private Mapper reporterMapper;
	private int numInserted;
	private int numRawBioAssays;
	private NumberFormat numberFormat;
	private boolean nullIfException;
	private boolean verifyColumns;
	private boolean nullIfMissingReporter;
	private boolean useSmartFeatureMismatchHandling;
	
	/**
		Create a FlatFileParser that can parse Illumina data files:
		<ul>
		<li>Data splitter: (,|\t)
		<li>Header regexp: (.+)=(.*?),*
		<li>Data header: TargetID(,|\t).*
		</ul>
		NOTE! To begin with we support both comma and tab as column splitter but
		later on (in {@link #isImportable(FlatFileParser)}) when we know which one is actually
		used, we change this in the parser. We need to do this since numbers may
		use comma as decimal separator.
	*/
	@Override
	protected FlatFileParser getInitializedFlatFileParser() 
		throws BaseException
	{
		String separator = "(,|\\t)";
		FlatFileParser ffp = new FlatFileParser();
		ffp.setDataSplitterRegexp(Pattern.compile(separator));
		ffp.setDataHeaderRegexp(Pattern.compile("TargetID"+separator + ".*"));
		ffp.setHeaderRegexp(Pattern.compile("(.+)=(.*?),*"));
		return ffp;
	}
	/**
		@return As specified by job parameter or "dot" if not
	*/
	@Override
	protected String getDecimalSeparator()
	{
		String separator = super.getDecimalSeparator();
		if (separator == null) separator = "dot";
		return separator;
	}

	/**
		Check that the first line contains the text "Illumina"
		@return TRUE if the first line contains "Illumina", FALSE
			otherwise
	*/
	@Override
	protected boolean isImportable(FlatFileParser ffp)
	{
		String firstLine = ffp.getLineCount() >= 1 ? ffp.getLine(0).line() : null;
		boolean isIllumina = firstLine != null && firstLine.contains("Illumina");
		if (isIllumina)
		{
			String separator = ",";
			FlatFileParser.Line lastLine = ffp.getLine(ffp.getLineCount()-1);
			if (lastLine.type() == LineType.DATA_HEADER)
			{
				int firstTab = lastLine.line().indexOf("\t");
				if (firstTab > 0 && firstTab < lastLine.line().indexOf(","))
				{
					separator = "\\t";
				}
				ffp.setDataSplitterRegexp(Pattern.compile(separator));
			}
		}
		return isIllumina;
	}
	
	@SuppressWarnings("unchecked")
	@Override
	protected void begin(FlatFileParser ffp) 
		throws BaseException
	{
		super.begin(ffp);
		this.ffp = ffp;
		this.illumina = RawDataTypes.getSafeRawDataType("illumina");
		this.numberFormat = ffp.getDefaultNumberFormat();
		this.headerLines = new LinkedList<Line>();
		this.experiment = (Experiment)job.getValue("experiment");
		this.bioAssays = (List<DerivedBioAssay>)job.getValues("bioAssays");
		this.design = (ArrayDesign)job.getValue("arrayDesign");
		this.protocol = (Protocol)job.getValue("protocol");
		this.software = (Software)job.getValue("software");
		
		// Feature identification
		try
		{
			String fiTemp = (String)job.getValue("featureIdentification");
			fiMethod = FeatureIdentificationMethod.valueOf(fiTemp);
		}
		catch (Exception ex)
		{
			fiMethod = null;
		}
		// Setup error handling
		this.nullIfException = "null".equals(getErrorOption("numberFormatError"));
		this.verifyColumns = "fail".equals(getErrorOption("invalidColumnsError"));
		if (design == null)
		{
			String method = getErrorOption("missingReporterError");
			this.nullIfMissingReporter = "null".equals(method);
			if (!nullIfMissingReporter && method != null)
			{
				addErrorHandler(ItemNotFoundException.class, new SimpleErrorHandler("skip".equals(method)));
			}
		}
		else
		{
			String method = (String)job.getValue("featureMismatchError");
			this.useSmartFeatureMismatchHandling = "smart".equals(method);
			if (method != null)
			{
				addErrorHandler(ItemNotFoundException.class, new SimpleErrorHandler("skip".equals(method)));
			}
			
		}
	}

	/**
		Copy headers so that we can set them on the raw bioassays in
		beginData()
	*/
	@Override
	protected void handleHeader(Line line) 
		throws BaseException
	{
		super.handleHeader(line);
		if (line.name() != null && line.value() != null)
		{
			headerLines.add(line);
		}
	}
	
	/**
		Check column headers and map them to raw bioassays.
		Create raw bioassays. Initialise column <code>Mapper</code>:s.
	*/
	@Override
	protected void beginData()
	{
		File rawDataFile = (File)job.getValue("file");
		this.dc = sc.newDbControl();
		try
		{
			this.rawBioAssays = extractAndCreateRawBioAssays(dc, ffp.getColumnHeaders(), verifyColumns, rawDataFile);
			// We must commit to be able to insert raw data
			// In case of a later failure the end() method should delete the raw bioassays again
			dc.commit();
		}
		finally
		{
			if (dc != null) dc.close();
		}
		this.numRawBioAssays = rawBioAssays.size();
		this.reporterMapper = new ColumnMapper(ffp.getColumnHeaderIndex("TargetID"), "Target ID");
		this.dc = sc.newDbControl();
		
		// Create RawDataBatcher and Mapper for each raw bioassay
		this.holders = new ArrayList<BatchAndMapHolder>(numRawBioAssays);
		for (RawBioAssay rba : rawBioAssays)
		{
			// Need to reload raw bioassay with current DbControl
			rba = RawBioAssay.getById(dc, rba.getId());
			RawDataBatcher batcher = rba.getRawDataBatcher(fiMethod, job.getJob());
			batcher.useNullIfReporterNotFound(nullIfMissingReporter);
			batcher.acceptInsertIfNullReporterOnFeature(useSmartFeatureMismatchHandling);
			batcher.skipInsertIfMissingFeature(useSmartFeatureMismatchHandling);
			holders.add(new BatchAndMapHolder(batcher, illumina, ffp));
		}
	}
	
	@Override
	protected void handleData(Data data) 
		throws BaseException
	{
		String externalId = reporterMapper.getValue(data);
		int lineNo = data.dataLineNo();
		for (BatchAndMapHolder holder : holders)
		{
			RawData raw = holder.batcher.newRawData();
			raw.setPosition(lineNo);
			raw.setBlock(1);
			raw.setColumn(1);
			raw.setRow(lineNo);
			for (Map.Entry<RawDataProperty, Mapper> entry : holder.mappers.entrySet())
			{
				RawDataProperty ep = entry.getKey();
				Mapper m = entry.getValue();
				raw.setExtended(ep.getName(), ep.parseString(m.getValue(data), numberFormat, nullIfException));
			}
			holder.batcher.insert(raw, externalId, externalId);
		}
		numInserted++;
	}
	
	/**
		If successful:
		<ul>
		<li>Close batchers
		<li>Associate with experiment if there is one
		<li>Commit
		</ul>
		If not successful:
		<ul>
		<li>Delete raw bioassays that has been created
		<li>Rollback
		</ul>
	*/
	@Override
	protected void end(boolean success)
		throws BaseException
	{
		try
		{
			if (success)
			{
				if (experiment != null) experiment = Experiment.getById(dc, experiment.getId());
				DataFileType rawDataType = DataFileType.getByExternalId(dc, DataFileType.GENERIC_RAW_DATA);
				for (BatchAndMapHolder holder : holders)
				{
					holder.batcher.close();
					if (experiment != null)
					{
						experiment.addRawBioAssay(holder.rba);
					}
					holder.rba.getFileSet().getMember(rawDataType).setValid(true, null);
				}
				dc.commit();
			}
		}
		catch (RuntimeException ex)
		{
			success = false;
			throw ex;
		}
		finally
		{
			if (dc != null) dc.close();
			if (!success && rawBioAssays != null)
			{
				Trashcan.delete(sc, rawBioAssays, true, null);				
			}
			super.end(success);
		}
	}
	
	@Override
	protected String getSuccessMessage(int skippedLines)
	{
		StringBuilder msg = new StringBuilder();
		// All batchers should have the same number of skipped/accepted spots...
		RawDataBatcher batcher = holders.get(0).batcher;
		msg.append(batcher.getTotalInsertCount()).append(" spot(s) inserted");
		int numReportersNotFound = batcher.getNumReportersNotFound();
		int numSkippedMissingFeature = batcher.getNumSkippedMissingFeature();
		int numAcceptedWithNullReporterOnFeature = batcher.getNumAcceptedWithNullReporterOnFeature();
		if (numReportersNotFound > 0)
		{
			msg.append("; ").append(numReportersNotFound).append(" spot(s) could not find a reporter");
		}
		if (numAcceptedWithNullReporterOnFeature > 0)
		{
			msg.append("; ").append(numAcceptedWithNullReporterOnFeature);
			msg.append(" spot(s) accepted with null reporter on feature");
		}
		if (numSkippedMissingFeature > 0)
		{
			msg.append("; ").append(numSkippedMissingFeature);
			msg.append(" spot(s) skipped due to missing features");
		}
		if (skippedLines > 0) 
		{
			msg.append("; ").append(skippedLines).append(" spot(s) skipped due to errors");
		}
		return msg.toString();		
	}
	// -------------------------------------------

	/**
		A column header must be like: extendedpropertyname-arrayname
	*/
	private static final Pattern splitHeader = Pattern.compile("(.+)\\-(.+)");
	
	/**
		Extract array names and raw data property names from the column headers.
		Create a raw bioassay for each array and set headers, scan, software
		and protocol. Optionally, verify that all arrays
		have the same data columns. Duplicate columns are always reported as an
		error.
	*/
	private List<RawBioAssay> extractAndCreateRawBioAssays(DbControl dc, List<String> headers, 
		boolean verifyColumns, File rawDataFile)
	{
		// Holds the created raw bioassays
		List<RawBioAssay> list = new LinkedList<RawBioAssay>();
		// Load the generic raw data file type
		DataFileType rawDataType = DataFileType.getByExternalId(dc, DataFileType.GENERIC_RAW_DATA);
		// Maps: array name -> set of extended property names
		Map<String, Set<String>> arrayNames = new HashMap<String, Set<String>>();
		// All extended property mames found
		Set<String> propertyNames = new HashSet<String>();
		for (String header : headers)
		{
			Matcher m = splitHeader.matcher(header);
			if (m.matches())
			{
				String propertyName = m.group(1);
				String arrayName = m.group(2);
				String lastInName = arrayName.substring(arrayName.length()-1);
				int arrayPosition = Coordinate.alphaToNumeric(lastInName);
				if (arrayPosition <= 0) arrayPosition = 1;
				// Check if raw data property exists
				if (illumina.getProperty(propertyName) == null)
				{
					if (verifyColumns)
					{
						throw new InvalidDataException("Unknown data column for array '" + 
								arrayName + "': " + propertyName);
					}
				}
				else
				{
					// Create raw bioassay if we havn't done so already
					if (!arrayNames.containsKey(arrayName))
					{
						arrayNames.put(arrayName, new HashSet<String>());
						Platform generic = Platform.getByExternalId(dc, Platform.GENERIC);
						RawBioAssay rba = RawBioAssay.getNew(dc, generic, illumina);
						rba.setName(arrayName);
						
						DerivedBioAssay bioAssay = null;
						Extract extract = null;
						if (bioAssays != null)
						{
							// Find derived bioassay and physical bioassay based on the array name
							bioAssay = findBioAssay(bioAssays, arrayName);
							if (bioAssay != null)
							{
								bioAssay = DerivedBioAssay.getById(dc, bioAssay.getId());
								// Find extract on the given array position.
								List<PhysicalBioAssay> physicalBioAssays = bioAssay.getPhysicalBioAssays().list(dc);
								if (physicalBioAssays.size() == 1)
								{
									List<BioMaterialEventSource> sources = physicalBioAssays.get(0).getCreationEvent().getEventSources(dc, arrayPosition);
									// There must be exactly one source
									if (sources.size() == 1) extract = (Extract)sources.get(0).getBioMaterial();
								}
							}
						}
						if (bioAssay != null) rba.setParentBioAssay(bioAssay);
						if (extract != null) rba.setParentExtract(extract);
						if (design != null) rba.setArrayDesign(design);
						if (protocol != null) rba.setProtocol(protocol);
						if (software != null) rba.setSoftware(software);
						if (rawDataFile != null) rba.getFileSet().setMember(rawDataFile, rawDataType);
						for (Line headerLine : headerLines)
						{
							rba.setHeader(headerLine.name(), headerLine.value());
						}
						list.add(rba);
						dc.saveItem(rba);
					}
					// Never allow duplicate data columns
					if (!arrayNames.get(arrayName).add(propertyName))
					{
						throw new InvalidDataException("Duplicate data column for array '" + arrayName + "': " + propertyName);
					}
					propertyNames.add(propertyName);
				}
			}
		}

		// Check that the same columns are included for all arrays
		if (verifyColumns)
		{
			for (Map.Entry<String, Set<String>> entry : arrayNames.entrySet())
			{
				String arrayName = entry.getKey();
				Set<String> properties = entry.getValue();
				for (String propertyName : propertyNames)
				{
					if (!properties.contains(propertyName))
					{
						throw new InvalidDataException("Missing data column for array '" + arrayName + "': " + propertyName);
					}
				}
			}
		}
		return list;
	}
	
	private DerivedBioAssay findBioAssay(List<DerivedBioAssay> bioAssays, String name)
	{
		if (bioAssays.size() == 1) return bioAssays.get(0);
		DerivedBioAssay bestMatch = null;
		StringMatcher matcher = new StringMatcher();
		double bestScore = 0;
		for (DerivedBioAssay dba : bioAssays)
		{
			double score = matcher.getScore(name, dba.getName());
			if (score > bestScore)
			{
				bestScore = score;
				bestMatch = dba;
			}
		}
		return bestMatch;
	}
	
	private RequestInformation getConfigureJobParameters(GuiContext context)
	{
		DbControl dc = null;
		try
		{
			if (configureJob == null)
			{
				List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
				
				parameters.add(fileParameter);
				parameters.add(Parameters.charsetParameter(null, null, null));
				parameters.add(Parameters.decimalSeparatorParameter(null, null, 
						(String)job.getValue(Parameters.DECIMAL_SEPARATOR_PARAMETER)));

				// parameters for scan, protocol, software and array design
				dc = sc.newDbControl();
				List<Protocol> protocols = getItems(dc, Protocol.getQuery(), 
						Restrictions.eq(
							Hql.property("itemSubtype.id"), 
							Expressions.integer(SystemItems.getId(Protocol.FEATURE_EXTRACTION))
						)
					);
				List<Software> software = getItems(dc, Software.getQuery(), 
					Restrictions.eq(
							Hql.property("itemSubtype.id"), 
							Expressions.integer(SystemItems.getId(Software.FEATURE_EXTRACTION))
						)
					);
				List<ArrayDesign> designs = getItems(dc, ArrayDesign.getQuery(),
					Restrictions.gt(
							Hql.property("numDbFeatures"), 
							Expressions.integer(0)
						)
					);
				List<DerivedBioAssay> bioAssays = getItems(dc, DerivedBioAssay.getQuery(),
					Restrictions.eq(
							Hql.property("itemSubtype.id"), 
							Expressions.integer(SystemItems.getId(DerivedBioAssay.SCAN))
						)
					);

				
				boolean hasAssociations = 
					context.getItem() == Item.EXPERIMENT ||  
					protocols.size() > 0 || software.size() > 0 || designs.size() > 0 || bioAssays.size() > 0;
				
				if (hasAssociations)
				{
					parameters.add(associationsSection);
					if (context.getItem() == Item.EXPERIMENT)
					{
						// Include parameter for current experiment
						parameters.add(new PluginParameter<Experiment>(
								"experiment",
								"Experiment",
								"The imported raw bioassays will be included in this experiment.",
								new ItemParameterType<Experiment>(Experiment.class, null, true, 1, null)
							));
					}
					if (!bioAssays.isEmpty())
					{
						parameters.add(new PluginParameter<DerivedBioAssay>(
							"bioAssays", 
							"Scans", 
							"The scans that the data in the file is coming from. " +
							"Select each scan that is used in the file. The information " +
							"will be used to link each raw bioassay to the extract the " +
							"data is coming from. If more than one scan is used, the names of the " +
							"scans should be similar to the data headers in the file.", 
							new ItemParameterType<DerivedBioAssay>(DerivedBioAssay.class, null, false, 0, bioAssays)
						));
					}
					
					if (!designs.isEmpty())
					{
						parameters.add(new PluginParameter<ArrayDesign>(
							"arrayDesign",
							"Array design",
							"The imported raw bioassays will be linked to the selected array design.",
							new ItemParameterType<ArrayDesign>(ArrayDesign.class, null, false, 1, designs)
						));
						parameters.add(featureIdentificationParameter);
					}
					if (!protocols.isEmpty())
					{
						parameters.add(new PluginParameter<Protocol>(
							"protocol",
							"Protocol",
							"The imported raw bioassays will be linked to the selected protocol.",
							new ItemParameterType<Protocol>(Protocol.class, null, false, 1, protocols)
						));
					}
					if (!software.isEmpty())
					{
						parameters.add(new PluginParameter<Software>(
							"software",
							"Software",
							"The imported raw bioassays will be linked to the selected software.",
							new ItemParameterType<Software>(Software.class, null, false, 1, software)
						));
					}
				}
				
				// Error handling parameters
				parameters.add(errorSection);
				parameters.add(defaultErrorParameter);
				parameters.add(invalidColumnsErrorParameter);
				parameters.add(missingReporterErrorParameter);
				parameters.add(featureMismatchErrorParameter);
				parameters.add(numberFormatErrorParameter);
				
				configureJob = new RequestInformation
				(
					Request.COMMAND_CONFIGURE_JOB,
					"Select the file to import raw data from",
					"You must select a file to use for import. Most other options are optional.",
					parameters
				);
			}
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return configureJob;
	}
	
	/**
		Convert a raw data property to a column name.
	*/
	private static String getColumnName(RawDataProperty rdp, RawBioAssay rba)
	{
		return rdp.getName() + "-" + rba.getName();
	}
	
	/**
		Sort the items by name and add USE permission filter to the query.
	*/
	private <T extends BasicItem> List<T> getItems(DbControl dc, ItemQuery<T> query, Restriction... restrictions)
	{
		query.order(Orders.asc(Hql.property("name")));
		query.include(Include.MINE, Include.SHARED, Include.IN_PROJECT);
		query.setItemPermission(Permission.USE);
		if (restrictions != null && restrictions.length > 0)
		{
			query.restrict(Restrictions.and(restrictions));
		}
		return query.list(dc);
	}
	
	/**
		Holds a RawBioAssay and the RawDataBatcher to use for import and the
		Mapper for all it's data columns.
	*/
	private static class BatchAndMapHolder
	{
		private final RawBioAssay rba;
		private final RawDataBatcher batcher;
		private final Map<RawDataProperty, Mapper> mappers;
		
		private BatchAndMapHolder(RawDataBatcher batcher, RawDataType illumina, FlatFileParser ffp)
		{
			this.rba = batcher.getRawBioAssay();
			this.batcher = batcher;
			this.mappers = new HashMap<RawDataProperty, Mapper>();
			createMappers(illumina, ffp);
		}
		
		private void createMappers(RawDataType illumina, FlatFileParser ffp)
		{
			for (RawDataProperty rawProperty : illumina.getProperties())
			{
				String colName = getColumnName(rawProperty, rba);
				Integer colIndex = ffp.getColumnHeaderIndex(colName);
				if (colIndex != null)
				{
					mappers.put(rawProperty, new ColumnMapper(colIndex, colName));
				}
			}
		}
	}
	
}
