/*
	$Id: TestBfs.java 5221 2010-01-22 08:26:41Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.Arrays;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.bfs.AnnotationModel;
import net.sf.basedb.util.bfs.AnnotationParser;
import net.sf.basedb.util.bfs.AnnotationWriter;
import net.sf.basedb.util.bfs.DataParser;
import net.sf.basedb.util.bfs.DataWriter;
import net.sf.basedb.util.bfs.MatrixModel;
import net.sf.basedb.util.bfs.MetadataModel;
import net.sf.basedb.util.bfs.MetadataParser;
import net.sf.basedb.util.bfs.MetadataWriter;
import net.sf.basedb.util.bfs.MetadataModel.Section;

public class TestBfs
{

	static boolean ok = true;
	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		write("++Testing BFS format");
		
		File metadata = test_create_temp_file("metadata");
		File metadata_copy = test_create_temp_file("metadata_copy");
		test_write_metadata(metadata);
		test_read_metadata(metadata, metadata_copy);
		test_read_metadata(metadata_copy, null);
		
		File annotations = test_create_temp_file("annotations");
		test_write_annotations(annotations);
		test_read_annotations(annotations);
		
		File data = test_create_temp_file("data");
		test_write_data(data);
		test_read_data(data);
		
		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
		
		write("++Testing BFS format "+(ok ? "OK" : "Failed")+"\n");
		return ok;
	}

	static File test_create_temp_file(String prefix)
	{
		File f = null;
		try
		{
			f = File.createTempFile(prefix, ".bfs.txt");
			f.deleteOnExit();
		}
		catch (Throwable t)
		{
			write("--Create BFS " + prefix + " FAILED");
			t.printStackTrace();
			ok = false;
		}
		return f;
	}
	
	static void test_write_metadata(File file)
	{
		if (file == null) return;
		MetadataWriter bfs = null;
		try
		{
			Writer out = new FileWriter(file);
			bfs = new MetadataWriter(out);
			bfs.setSubtype("test");
			
			// Test comments and empty lines
			bfs.bfsPrintComment("A comment line");
			bfs.bfsPrintEmptyLine(1);
			bfs.bfsPrintComment("A multiline comment\nis automatically generated");
			bfs.bfsPrintEmptyLine(1);

			// First section, a duplicate key is not allowed
			bfs.bfsPrintSection("one", true, true);
			bfs.bfsPrintValue("first", "The first value");
			bfs.bfsPrintValue("second", "The second value");
			try
			{
				bfs.bfsPrintValue("second", "Not allowed");
				throw new BaseException("BFSMetadataWriter did not detect duplicate value");
			}
			catch (IllegalArgumentException ex)
			{} // Expected
			
			// Numeric values and a vector
			bfs.bfsPrintValue("float", 8.0);
			bfs.bfsPrintValue("vector", 1.22, 2.59999, -1.0);			
			bfs.bfsPrintEmptyLine(1);
			
			// Second section allow duplicate keys
			bfs.bfsPrintSection("two", true, false);
			bfs.bfsPrintValue("float", 8.0);
			bfs.bfsPrintValue("float", 9.0);
			
			// A duplicate section is not allowed
			try
			{
				bfs.bfsPrintSection("two", true, false);
				throw new BaseException("BFSMetadataWriter did not detect duplicate section");
			}
			catch (IllegalArgumentException ex)
			{} // Expected
			
			// The 'files' section is a pre-defined section
			bfs.bfsPrintSection("files", true, true);
			bfs.bfsPrintValue("file-1", "abc123.txt");
			bfs.bfsPrintValue("file-2", "def456.png");
			
			bfs.flush();
			bfs.close();
			write("--Test BFS Metadata writer OK");
		}
		catch (Throwable ex)
		{
			write("--Test BFS Metadata writer FAILED: " + file);
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (bfs != null)
			{
				bfs.flush();
				bfs.close();
			}
		}
	}

	static void test_read_metadata(File file, File copyTo)
	{
		if (file == null) return;
		boolean log = !TestUtil.getSilent();
		if (log) write("  Parsing BFS Metadata: " + file);
		try
		{
			MetadataParser bfsParser = new MetadataParser();
			MetadataModel bfs = new MetadataModel();
			bfsParser.setInputStream(new FileInputStream(file));
			bfsParser.setFilename(file.getAbsolutePath());
			bfsParser.parse(bfs);
			
			if (log) 
			{
				write("  BFS Subtype: " + bfs.getSubtype());
				write("  Sections: " + bfs.getSectionCount());
				for (int i = 0; i < bfs.getSectionCount(); ++i)
				{
					Section s = bfs.getSection(i);
					write("  Section " + i + ": " + s.getName());
					for (int j = 0; j < s.getEntryCount(); ++j)
					{
						write("    " + s.getKey(j) + "=" + s.getValue(j));
					}
				}
			}
			if (!"test".equals(bfs.getSubtype()))
			{
				throw new BaseException("Unexpected BFS subtype: " + bfs.getSubtype());
			}
			
			if (bfs.getSectionCount() != 3)
			{
				throw new BaseException("Unexpected number of sections: " + bfs.getSectionCount());				
			}
			Section one = bfs.getSection(0);
			if (!"one".equals(one.getName()))
			{
				throw new BaseException("Unexpected name of first section: " + one.getName());				
			}
			if (one.getEntryCount() != 4)
			{
				throw new BaseException("Unexpected number of values in section 'one': " + one.getEntryCount());				
			}
			if (!"first".equals(one.getKey(0)))
			{
				throw new BaseException("Unexpected key in section 'one', index 0: " + one.getKey(0));				
			}
			if (!"The first value".equals(one.getValue(0)))
			{
				throw new BaseException("Unexpected value in section 'one', index 0: " + one.getValue(0));				
			}
			if (!"The second value".equals(one.getValue("second")))
			{
				throw new BaseException("Unexpected value in section 'one', index 'second': " + one.getValue("second"));				
			}
			String[] vector = one.getValues("vector");
			if (vector.length != 3)
			{
				throw new BaseException("Unexpected value in section 'one', index 'vector': " + one.getValue("vector"));
			}

			Section two = bfs.getSection("two");
			if (two == null)
			{
				throw new BaseException("Can't find section: two");
			}
			if (two.getEntryCount() != 2)
			{
				throw new BaseException("Unexpected number of values in section 'two': " + two.getEntryCount());				
			}
			if (!"float".equals(two.getKey(0)))
			{
				throw new BaseException("Unexpected key in section 'two', index 0: " + two.getKey(0));				
			}
			if (!"float".equals(two.getKey(1)))
			{
				throw new BaseException("Unexpected key in section 'two', index 1: " + two.getKey(1));				
			}
			
			if (bfs.getFileCount() != 2)
			{
				throw new BaseException("Unexpected number of file: " + bfs.getFileCount());				
			}
			if (!"abc123.txt".equals(bfs.getFile(0)))
			{
				throw new BaseException("Unexpected file, index 0: " + bfs.getFile(0));				
			}
			if (!"def456.png".equals(bfs.getFile("file-2")))
			{
				throw new BaseException("Unexpected file, index 'file-2': " + bfs.getFile("file-2"));				
			}
			
			if (copyTo != null)
			{
				bfs.print(new OutputStreamWriter(new FileOutputStream(copyTo), "UTF-8"));
			}
			
			write("--Test BFS Metadata parser OK");
		}
		catch (Throwable ex)
		{
			write("--Test BFS Metadata parser FAILED: " + file);
			ex.printStackTrace();
			ok = false;
		}
	}

	static void test_write_annotations(File file)
	{
		if (file == null) return;
		AnnotationWriter bfs = null;
		try
		{
			Writer out = new FileWriter(file);
			bfs = new AnnotationWriter(out);
			
			// Invalid headers with duplicates
			try
			{
				bfs.bfsPrintHeaders("name", "description", "name");
				throw new BaseException("AnnotationWriter did not detect duplicate column name");
			}
			catch (IllegalArgumentException ex)
			{} // Expected
			
			// Valid headers
			bfs.bfsPrintHeaders("name", "decription", "size");

			// Two valid lines with data
			bfs.bfsPrintData(1, "item-1", "The first item", 55.3);
			bfs.bfsPrintData(2, "item-2", "The second item has a <\t> (tab) and <\n> (linebreak)", -44);
			
			// A duplicate id
			try
			{
				bfs.bfsPrintData(1, "duplicate", "", null);
				throw new BaseException("AnnotationWriter did not detect duplicate row id");
			}
			catch (IllegalArgumentException ex)
			{} // Expected

			// Incorrect number of columns
			try
			{
				bfs.bfsPrintData(3, "item-3", "The third item is missing size");
				throw new BaseException("AnnotationWriter did not detect incorrect number of columns");
			}
			catch (IllegalArgumentException ex)
			{} // Expected
			
			bfs.flush();
			bfs.close();
			write("--Test BFS Annotation writer OK");
		}
		catch (Throwable ex)
		{
			write("--Test BFS Annotation writer FAILED: " + file);
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (bfs != null)
			{
				bfs.flush();
				bfs.close();
			}
		}
	}	
	
	static void test_read_annotations(File file)
	{
		if (file == null) return;
		final boolean log = !TestUtil.getSilent();
		if (log) write("  Parsing BFS Annotations: " + file);
		try
		{
			AnnotationParser bfsParser = new AnnotationParser();
			AnnotationModel bfs = new AnnotationModel();
			bfsParser.setInputStream(new FileInputStream(file));
			bfsParser.setFilename(file.getAbsolutePath());
			bfsParser.parse(bfs);

			String[] headers = bfs.getHeaders();
			int rowCount = bfs.getRowCount();
			int columnCount = bfs.getColumnCount();
			
			if (log)
			{
				write("  Rows: " + rowCount + "; columns: " + columnCount);
				write("  Header: " + 
						Values.getString(Arrays.asList(headers), "\t", false));
				
				for (int r = 0; r < rowCount; r++)
				{
					int rowId = bfs.getRowId(r);
					String[] row = bfs.getRowById(rowId);
					write("  Line " + r + ": " + 
							Values.getString(Arrays.asList(row), "\t", false));
					if (row.length != columnCount)
					{
						throw new BaseException("Expected " + columnCount + 
								" data columns, not " + row.length);
					}
				}
			}
			
			if (columnCount != 4)
			{
				throw new BaseException("Expected 4 header columns, not " + headers.length);
			}
			
			if (rowCount != 2)
			{
				throw new BaseException("Expected 2 lines with data, not " + rowCount);
			}
			
			if (bfs.getRowId(0) != 1)
			{
				throw new BaseException("Unexpected row id at line 0: " + bfs.getRowId(0));
			}
			if (bfs.getRowIndex(2) != 1)
			{
				throw new BaseException("Unexpected row indx for row id '2': " + bfs.getRowIndex(2));
			}
			
			write("--Test BFS Annotation parser OK");
		}
		catch (Throwable ex)
		{
			write("--Test BFS Annotation parser FAILED: " + file);
			ex.printStackTrace();
			ok = false;
		}
	}
	
	static void test_write_data(File file)
	{
		if (file == null) return;
		DataWriter bfs = null;
		try
		{
			Writer out = new FileWriter(file);
			bfs = new DataWriter(out);
			
			bfs.setColumnCount(4);
			

			// Two valid lines with data
			bfs.bfsPrintData(1, "item-1", "The first item", 55.3);
			bfs.bfsPrintData(2, "item-2", "The second item has a <\t> (tab) and <\n> (linebreak)", -44);
			
			// Incorrect number of columns
			try
			{
				bfs.bfsPrintData(3, "item-3", "The third line is missing a column");
				throw new BaseException("DataWriter did not detect incorrect number of columns");
			}
			catch (IllegalArgumentException ex)
			{} // Expected
			
			bfs.flush();
			bfs.close();
			write("--Test BFS Data writer OK");
		}
		catch (Throwable ex)
		{
			write("--Test BFS Data writer FAILED: " + file);
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (bfs != null)
			{
				bfs.flush();
				bfs.close();
			}
		}
	}

	static void test_read_data(File file)
	{
		if (file == null) return;
		final boolean log = !TestUtil.getSilent();
		if (log) write("  Parsing BFS Data: " + file);
		try
		{
			DataParser bfsParser = new DataParser();
			MatrixModel bfs = new MatrixModel();
			bfsParser.setInputStream(new FileInputStream(file));
			bfsParser.setFilename(file.getAbsolutePath());
			bfsParser.parse(bfs);

			int rowCount = bfs.getRowCount();
			int columnCount = bfs.getColumnCount();
			
			if (log)
			{
				write("  Rows: " + rowCount + "; columns: " + columnCount);
				
				for (int r = 0; r < rowCount; r++)
				{
					String[] row = bfs.getRowByIndex(r);
					write("  Line " + r + ": " + 
							Values.getString(Arrays.asList(row), "\t", false));
					if (row.length != columnCount)
					{
						throw new BaseException("Expected " + columnCount + 
								" data columns, not " + row.length);
					}
				}
			}
			
			if (columnCount != 4)
			{
				throw new BaseException("Expected 4 data columns, not " + columnCount);
			}
			
			if (rowCount != 2)
			{
				throw new BaseException("Expected 2 lines with data, not " + rowCount);
			}
						
			write("--Test BFS Data parser OK");
		}
		catch (Throwable ex)
		{
			write("--Test BFS Data parser FAILED: " + file);
			ex.printStackTrace();
			ok = false;
		}
	}

	
	static void write(String message)
	{
		System.out.println(message);
	}
	
}

